@@ -9,11 +9,13 @@ if _IS_PYTHON_2:
     from urllib2 import Request as URLRequest
     from urllib2 import urlopen
     from urllib import urlencode
+    from urllib import quote as urlescape
 else:
     from urllib.error import HTTPError
     from urllib.request import Request as URLRequest
     from urllib.request import urlopen
     from urllib.parse import urlencode
+    from urllib.parse import quote as urlescape
 
 import uuid
 import json
@@ -21,6 +23,7 @@ import traceback
 import cmd
 import zlib
 import base64
+import time
 from functools import wraps
 
 from .Sensor import Sensor
@@ -30,9 +33,12 @@ from .utils import GET
 from .utils import POST
 from .utils import DELETE
 
+from .Jobs import Job
+
 from limacharlie import GLOBAL_OID
 from limacharlie import GLOBAL_UID
 from limacharlie import GLOBAL_API_KEY
+from limacharlie import _getEnvironmentCreds
 
 ROOT_URL = 'https://api.limacharlie.io'
 API_VERSION = 'v1'
@@ -40,36 +46,48 @@ API_VERSION = 'v1'
 API_TO_JWT_URL = 'https://app.limacharlie.io/jwt'
 
 HTTP_UNAUTHORIZED = 401
+HTTP_TOO_MANY_REQUESTS = 429
+HTTP_GATEWAY_TIMEOUT = 504
+HTTP_OK = 200
 
 class Manager( object ):
     '''General interface to a limacharlie.io Organization.'''
 
-    def __init__( self, oid, secret_api_key, inv_id = None, print_debug_fn = None, is_interactive = False, extra_params = {}, jwt = None, uid = None, onRefreshAuth = None ):
+    def __init__( self, oid = None, secret_api_key = None, environment = None, inv_id = None, print_debug_fn = None, is_interactive = False, extra_params = {}, jwt = None, uid = None, onRefreshAuth = None, isRetryQuotaErrors = False ):
         '''Create a session manager for interaction with limacharlie.io, much of the Python API relies on this object.
 
         Args:
-            oid (str): a limacharlie.io organization ID.
-            secret_api_key (str): an API key for the organization, as provided by limacharlie.io.
+            oid (str): a limacharlie.io organization ID, default environment if unset.
+            secret_api_key (str): an API key for the organization, as provided by limacharlie.io, default environment if unset.
+            environment (str): an environment name as defined in "limacharlie login" to use.
             inv_id (str): an investigation ID that will be used/propagated to other APIs using this Manager instance.
             print_debug_fn (function(message)): a callback function that will receive detailed debug messages.
             is_interactive (bool): if True, the manager will provide a root investigation and Spout so that tasks sent to Sensors can be tracked in realtime automatically; requires an inv_id to be set.
             extra_params (dict): optional key / values passed to interactive spout.
             jwt (str): optionally specify a single JWT to use for authentication.
-            uid (str): a limacharlie.io user ID, if present authentication will be based on it instead of organization ID.
+            uid (str): a limacharlie.io user ID, if present authentication will be based on it instead of organization ID, set to False to override the current environment.
             onRefreshAuth (func): if provided, function is called whenever a JWT would be refreshed using the API key.
+            isRetryQuotaErrors (bool): if True, the Manager will attempt to retry queries when it gets an out-of-quota error (HTTP 429).
         '''
-        # If no creds were provided, use the global ones.
-        if oid is None:
-            if GLOBAL_OID is None:
-                raise Exception( 'LimaCharlie "default" environment not set, please use "limacharlie login".' )
-            oid = GLOBAL_OID
-        if uid is None:
-            if GLOBAL_UID is not None:
-                uid = GLOBAL_UID
-        if secret_api_key is None:
-            if GLOBAL_API_KEY is None and jwt is None:
-                raise Exception( 'LimaCharlie "default" environment not set, please use "limacharlie login".' )
-            secret_api_key = GLOBAL_API_KEY
+        # If an environment is specified, try to get its creds.
+        if environment is not None:
+            oid, uid, secret_api_key = _getEnvironmentCreds( environment )
+            if secret_api_key is None or ( oid is None and uid is None ):
+                raise LcApiException( 'LimaCharlie environment not configured, use "limacharlie login".')
+        else:
+            # Otherwise, try to take the values in parameter. But if
+            # they are not present, use the GLOBAL values.
+            if oid is None:
+                if GLOBAL_OID is None:
+                    raise LcApiException( 'LimaCharlie "default" environment not set, please use "limacharlie login".' )
+                oid = GLOBAL_OID
+            if uid is None:
+                if GLOBAL_UID is not None:
+                    uid = GLOBAL_UID
+            if secret_api_key is None and jwt is None:
+                if GLOBAL_API_KEY is None:
+                    raise LcApiException( 'LimaCharlie "default" environment not set, please use "limacharlie login".' )
+                secret_api_key = GLOBAL_API_KEY
 
         try:
             uuid.UUID( oid )
@@ -81,7 +99,7 @@ class Manager( object ):
             if jwt is None:
                 raise LcApiException( 'Invalid secret API key, should be in UUID format.' )
         self._oid = oid
-        self._uid = uid
+        self._uid = uid if uid else None
         self._onRefreshAuth = onRefreshAuth
         self._secret_api_key = secret_api_key
         self._jwt = jwt
@@ -91,13 +109,17 @@ class Manager( object ):
         self._spout = None
         self._is_interactive = is_interactive
         self._extra_params = extra_params
+        self._isRetryQuotaErrors = isRetryQuotaErrors
         if self._is_interactive:
             if not self._inv_id:
                 raise LcApiException( 'Investigation ID must be set for interactive mode to be enabled.' )
             self._refreshSpout()
 
-    def _unwrap( self, data ):
-        return json.loads( zlib.decompress( base64.b64decode( data ), 16 + zlib.MAX_WBITS ).decode() )
+    def _unwrap( self, data, isRaw = False ):
+        if isRaw:
+            return zlib.decompress( base64.b64decode( data ), 16 + zlib.MAX_WBITS )
+        else:
+            return json.loads( zlib.decompress( base64.b64decode( data ), 16 + zlib.MAX_WBITS ).decode() )
 
     def _refreshSpout( self ):
         if not self._is_interactive:
@@ -111,15 +133,17 @@ class Manager( object ):
         if self._debug is not None:
             self._debug( msg )
 
-    def _refreshJWT( self ):
+    def _refreshJWT( self, expiry = None ):
         try:
             if self._secret_api_key is None:
                 raise Exception( 'No API key set' )
             authData = { "secret" : self._secret_api_key }
             if self._uid is not None:
                 authData[ 'uid' ] = self._uid
-            else:
+            if self._oid is not None:
                 authData[ 'oid' ] = self._oid
+            if expiry is not None:
+                authData[ 'expiry' ] = int( expiry )
             request = URLRequest( API_TO_JWT_URL,
                                   urlencode( authData ).encode() )
             request.get_method = lambda: "POST"
@@ -128,7 +152,7 @@ class Manager( object ):
             u.close()
         except Exception as e:
             self._jwt = None
-            raise LcApiException( 'Failed to get JWT from API key: %s' % e )
+            raise LcApiException( 'Failed to get JWT from API key oid=%s uid=%s: %s' % ( self._oid, self._uid, e, ) )
 
     def _restCall( self, url, verb, params, altRoot = None, queryParams = None, rawBody = None, contentType = None, isNoAuth = False ):
         try:
@@ -174,23 +198,53 @@ class Manager( object ):
 
         return ret
 
-    def _apiCall( self, url, verb, params = {}, altRoot = None, queryParams = None, rawBody = None, contentType = None, isNoAuth = False ):
+    def _apiCall( self, url, verb, params = {}, altRoot = None, queryParams = None, rawBody = None, contentType = None, isNoAuth = False, nMaxTotalRetries = 3 ):
+        hasAuthRefreshed = False
+        nRetries = 0
+
+        # If no JWT is ready, prime it.
         if not isNoAuth and self._jwt is None:
+            hasAuthRefreshed = True
             if self._onRefreshAuth is not None:
                 self._onRefreshAuth( self )
             else:
                 self._refreshJWT()
 
-        code, data = self._restCall( url, verb, params, altRoot = altRoot, queryParams = queryParams, rawBody = rawBody, contentType = contentType, isNoAuth = isNoAuth )
+        while nRetries < nMaxTotalRetries:
+            nRetries += 1
 
-        if code == HTTP_UNAUTHORIZED and not isNoAuth:
-            if self._onRefreshAuth is not None:
-                self._onRefreshAuth( self )
-            else:
-                self._refreshJWT()
-            code, data = self._restCall( url, verb, params, altRoot = altRoot, queryParams = queryParams, rawBody = rawBody, contentType = contentType )
+            code, data = self._restCall( url, verb, params, altRoot = altRoot, queryParams = queryParams, rawBody = rawBody, contentType = contentType, isNoAuth = isNoAuth )
 
-        if 200 != code:
+            if code == HTTP_UNAUTHORIZED:
+                if hasAuthRefreshed:
+                    # We already renewed the JWT once.
+                    break
+                elif not isNoAuth:
+                    # Do our one JWT renew attempt.
+                    hasAuthRefreshed = True
+                    if self._onRefreshAuth is not None:
+                        self._onRefreshAuth( self )
+                    else:
+                        self._refreshJWT()
+                    continue
+                else:
+                    # Auth failed, can't renew.
+                    break
+
+            if code == HTTP_TOO_MANY_REQUESTS and self._isRetryQuotaErrors:
+                # Out of quota, wait a bit and retry.
+                time.sleep( 10 )
+                continue
+
+            if code == HTTP_GATEWAY_TIMEOUT:
+                # The API gateway timed out talking to the
+                # backend, we'll give it another shot.
+                continue
+
+            # Some other status code, including 200, so we're done.
+            break
+
+        if code != HTTP_OK:
             raise LcApiException( 'Api failure (%s): %s' % ( code, str( data ) ) )
 
         return data
@@ -302,37 +356,48 @@ class Manager( object ):
         return s
 
     def sensors( self, inv_id = None, is_next = False ):
-        '''Get the list of all Sensors in the Organization.
+        '''Gets all Sensors in the Organization.
 
         The sensors may or may not be online.
 
         Args:
             inv_id (str): investigation ID to add to all actions done using these objects.
-            is_next (bool): if set to True, will get the next slice of Sensors (if previous call to .sensors() hit the maximum number returned).
+
+        Returns:
+            a generator of Sensor objects.
+        '''
+
+        continuationToken = None
+
+        while True:
+            params = {}
+
+            if continuationToken is not None:
+                params[ 'continuation_token' ] = continuationToken
+
+            resp = self._apiCall( 'sensors/%s' % self._oid, GET, queryParams = params )
+            if inv_id is None:
+                inv_id = self._inv_id
+
+            for s in resp[ 'sensors' ]:
+                yield self.sensor( s[ 'sid' ], inv_id )
+
+            continuationToken = resp.get( 'continuation_token', None )
+            if continuationToken is None:
+                break
+
+    def sensorsWithTag( self, tag ):
+        '''Get a list of sensors that have the matching tag.
+
+        Args:
+            tag (str): a tag to look for.
 
         Returns:
             a list of Sensor objects.
         '''
 
-        if is_next and self._lastSensorListContinuationToken is None:
-            return None
-
-        params = {}
-        if is_next:
-            if self._lastSensorListContinuationToken is None:
-                return []
-            params[ 'continuation_token' ] = self._lastSensorListContinuationToken
-            self._lastSensorListContinuationToken = None
-
-        sensors = []
-        resp = self._apiCall( 'sensors/%s' % self._oid, GET, queryParams = params )
-        if inv_id is None:
-            inv_id = self._inv_id
-        for s in resp[ 'sensors' ]:
-            sensors.append( self.sensor( s[ 'sid' ], inv_id ) )
-        self._lastSensorListContinuationToken = resp.get( 'continuation_token', None )
-
-        return sensors
+        resp = self._apiCall( 'tags/%s/%s' % ( self._oid, urlescape( tag, '' ) ), GET, queryParams = {} )
+        return [ Sensor( self, sid ) for sid in resp.keys() ]
 
     def outputs( self ):
         '''Get the list of all Outputs configured for the Organization.
@@ -382,7 +447,7 @@ class Manager( object ):
         '''Get the Sensor objects for hosts matching a hostname expression.
 
         Args:
-            hostname_expr (str): hostname to look for, where '%' is a wildcard.
+            hostname_expr (str): hostname prefix to look for.
 
         Returns:
             a list of Sensor IDs matching the hostname expression.
@@ -456,6 +521,58 @@ class Manager( object ):
 
         return self._apiCall( 'rules/%s' % self._oid, POST, req )
 
+    def fps( self ):
+        '''Get the list of all False Positive rules for the Organization.
+
+        Returns:
+            a list of False Positive rules (JSON).
+        '''
+
+        req = {}
+
+        resp = self._apiCall( 'fp/%s' % self._oid, GET, queryParams = req )
+        return resp
+
+    def del_fp( self, name ):
+        '''Remove a False Positive rule from the Organization.
+
+        Args:
+            name (str): the name of the rule to remove.
+
+        Returns:
+            the REST API response (JSON).
+        '''
+
+        req = {
+            'name' : name,
+        }
+
+        return self._apiCall( 'fp/%s' % self._oid, DELETE, req )
+
+    def add_fp( self, name, rule, isReplace = False ):
+        '''Add a False Positive rule to the Organization.
+
+        For detailed explanation and possible rules parameters
+        see the official documentation, naming is the same as for the
+        REST interface.
+
+        Args:
+            name (str): name to give to the rule.
+            isReplace (boolean): if True, replace existing rule with the same name.
+            detection (dict): dictionary representing the False Positive rule content.
+
+        Returns:
+            the REST API response (JSON).
+        '''
+
+        req = {
+            'name' : name,
+            'is_replace' : 'true' if isReplace else 'false',
+            'rule' : json.dumps( rule ),
+        }
+
+        return self._apiCall( 'fp/%s' % self._oid, POST, req )
+
     def isInsightEnabled( self ):
         '''Check to see if Insight (retention) is enabled on this organization.
 
@@ -477,8 +594,9 @@ class Manager( object ):
             cat (str): return dects only from this category.
 
         Returns:
-            a list of detects.
+            a generator of detects.
         '''
+        cursor = '-'
         start = int( start )
         end = int( end )
         if limit is not None:
@@ -496,8 +614,18 @@ class Manager( object ):
         if cat is not None:
             req[ 'cat' ] = cat
 
-        data = self._apiCall( 'insight/%s/detections' % ( self._oid, ), GET, queryParams = req )
-        return self._unwrap( data[ 'detects' ] )
+        nReturned = 0
+        while cursor:
+            req[ 'cursor' ] = cursor
+            data = self._apiCall( 'insight/%s/detections' % ( self._oid, ), GET, queryParams = req )
+            cursor = data.get( 'next_cursor', None )
+            for detect in self._unwrap( data[ 'detects' ] ):
+                yield detect
+                nReturned += 1
+                if limit is not None and limit <= nReturned:
+                    break
+            if limit is not None and limit <= nReturned:
+                break
 
     def getObjectInformation( self, objType, objName, info, isCaseSensitive = True, isWithWildcards = False ):
         '''Get information about an object (indicator) using Insight (retention) data.
@@ -589,31 +717,57 @@ class Manager( object ):
         } )
         return data.get( 'sid', None )
 
-    def replicantRequest( self, replicantName, data, isSynchronous = False ):
-        '''Issue a request to a Replicant.
+    def getSensorsWithIp( self, ip, start, end ):
+        '''Get the list of sensor IDs that used the given IP during the time range.
 
         Args:
-            replicantName (str): the name of the Replicant to task.
-            data (dict): JSON data to send to the Replicant as a request.
-            isSynchronous (bool): if set to True, wait for data from the Replicant and return it.
+            ip (str): the IP address used.
+            start (int): beginning of the time range to look for.
+            end (int): end of the time range to look for.
+
         Returns:
-            Dict with general success, or data from Replicant if isSynchronous.
+            List of sid.
         '''
-        data = self._apiCall( 'replicant/%s/%s' % ( self._oid, replicantName ), POST, {
+        data = self._apiCall( 'ips/%s' % ( self._oid, ), GET, queryParams = {
+            'ip' : str( ip ),
+            'start' : int( start ),
+            'end' : int( end ),
+        } )
+        return data.get( 'sid', None )
+
+    def serviceRequest( self, serviceName, data, isAsynchronous = False ):
+        '''Issue a request to a Service.
+
+        Args:
+            serviceName (str): the name of the Service to task.
+            data (dict): JSON data to send to the Service as a request.
+            isAsynchronous (bool): if set to False, wait for data from the Service and return it.
+        Returns:
+            Dict with general success, or data from Service if isSynchronous.
+        '''
+        data = self._apiCall( 'service/%s/%s' % ( self._oid, serviceName ), POST, {
             'request_data' : base64.b64encode( json.dumps( data ).encode() ),
-            'is_async' : isSynchronous,
+            'is_async' : isAsynchronous,
         } )
         return data
 
-    def getAvailableReplicants( self ):
-        '''Get the list of Replicants currently available.
+    def replicantRequest( self, *args, **kwargs ):
+        # Maintained for backwards compatibility post rename replicant => service.
+        return self.serviceRequest( *args, **kwargs )
+
+    def getAvailableServices( self ):
+        '''Get the list of Services currently available.
 
         Returns:
-            List of Replicant names.
+            List of Service names.
         '''
-        data = self._apiCall( 'replicant/%s' % ( self._oid, ), GET )
+        data = self._apiCall( 'service/%s' % ( self._oid, ), GET )
         return data.get( 'replicants', None )
 
+    def getAvailableReplicants( self ):
+        # Maintained for backwards compatibility post rename replicant => service.
+        return self.getAvailableServices()
+
     def getOrgConfig( self, configName ):
         '''Get the value of a per-organization config.
 
@@ -773,6 +927,86 @@ class Manager( object ):
         } )
         return data
 
+    def getJobs( self, startTime, endTime, limit = None, sid = None ):
+        '''Get all the jobs in an organization in a time window.
+
+        Args:
+            startTime (int): second epoch of the start of the time window.
+            endTime (int): second epoch of the end of the time window.
+            limit (int): optional maximum number of jobs to return.
+            sid (str): optionally only return jobs that relate to this sensor ID.
+        Returns:
+            a Job object.
+        '''
+        params = {
+            'start' : startTime,
+            'end' : endTime,
+            'is_compressed' : 'true',
+            'with_data' : 'false',
+        }
+        if limit is not None:
+            params[ 'limit' ] = limit
+        if sid is not None:
+            params[ 'sid' ] = sid
+        data = self._apiCall( 'job/%s' % ( self._oid, ), GET, queryParams = params )
+        data = self._unwrap( data[ 'jobs' ] )
+        data = [ Job( self, job ) for jobId, job in data.items() ]
+        return data
+
+    def getJob( self, jobId ):
+        '''Get a specific job.
+
+        Args:
+            jobId (str): job ID of the job to get.
+        Returns:
+            a Job object.
+        '''
+        job = Job( self, { 'job_id' : jobId } )
+        job.update()
+        return job
+
+    def getApiKeys( self ):
+        '''Get the list of API keys in the organization.
+
+        '''
+        data = self._apiCall( 'orgs/%s/keys' % ( self._oid, ), GET, {} )
+        return data.get( 'api_keys', None )
+
+    def addApiKey( self, keyName, permissions = [] ):
+        '''Add an API key to an organization.
+
+        Args:
+            keyName (str): name of the key to add.
+            permissions (str[]): list of permissions for the key.
+        Returns:
+            the secret value of the new API key.
+        '''
+        data = self._apiCall( 'orgs/%s/keys' % ( self._oid, ), POST, {
+            'key_name' : keyName,
+            'perms' : ",".join( permissions ),
+        } )
+        return data
+
+    def removeApiKey( self, keyHash ):
+        '''Remove an API key from an organization.
+
+        Args:
+            keyHash (str): key hash of the key to remove.
+        '''
+        data = self._apiCall( 'orgs/%s/keys' % ( self._oid, ), DELETE, {
+            'key_hash' : keyHash,
+        } )
+        return data
+
+    def exportSensorList( self ):
+        '''Perform a bulk export of the entire sensor list.
+
+        Returns:
+            a dictionary of sensors with their information and tags.
+        '''
+        data = self._apiCall( 'export/%s/sensors' % ( self._oid, ), POST, {} )
+        return data
+
 def _eprint( msg ):
     sys.stderr.write( msg )
     sys.stderr.write( "\n" )
