@@ -1,5 +1,10 @@
-from gevent import monkey
-monkey.patch_all()
+try:
+    from gevent import monkey
+    monkey.patch_all()
+except monkey.MonkeyPatchWarning as e:
+    import sys
+    sys.stderr.write( "%s\n" % ( e, ) )
+    sys.stderr.flush()
 from gevent.lock import BoundedSemaphore
 from . import Manager
 from .utils import LcApiException
@@ -16,6 +21,15 @@ class Replay( object ):
     '''Interface to query historical sensor data in Insight with specific D&R rules.'''
 
     def __init__( self, manager, maxTimeWindow = ( 60 * 60 * 24 * 1 ), maxConcurrent = 10, isInteractive = False ):
+        '''Create a Replay manager object.
+
+        Args:
+            manager (limacharlie.Manager): manager providing authentication.
+            maxTimeWindow (int): number of seconds to split sensor data during analysis.
+            maxConcurrent (int): number of sensors windows to process in parallel.
+            isInteractive (bool): if True, display progress updates to standard out.
+        '''
+
         self._lc = manager
         self._apiURL = None
         self._maxTimeWindow = maxTimeWindow
@@ -26,6 +40,8 @@ class Replay( object ):
         self._statusMutex = BoundedSemaphore()
         self._queryPending = 0
         self._sensorPending = 0
+        self._limitEvent = None
+        self._imitEval = None
         self._queryStartedAt = None
         if self._isInteracive:
             self._queryStartedAt = time.time()
@@ -49,7 +65,7 @@ class Replay( object ):
 
         gevent.spawn_later( timing, self._reportStatus )
 
-    def scanHistoricalSensor( self, sid, startTime, endTime, ruleName = None, ruleContent = None, isRunTrace = False ):
+    def scanHistoricalSensor( self, sid, startTime, endTime, ruleName = None, ruleContent = None, isRunTrace = False, limitEvent = None, limitEval = None, isIgnoreState = False ):
         '''Scan a specific sensor's data with a D&R rule.
 
         Args:
@@ -59,6 +75,9 @@ class Replay( object ):
             ruleName (str): the name of an existing D&R rule to use.
             ruleContent (dict): D&R rule to use to scan, with a "detect" key and a "respond" key.
             isRunTrace (bool): if True, generate a trace of the evaluation.
+            limitEvent (int): approximately limit the number of events evaluated.
+            limitEval (int): approximately limit the number of rule evaluations.
+            isIgnoreState (bool): if True, parallelize processing of single sensors to increase performance but limit effectiveness of stateful detection.
 
         Returns:
             a dict containing results of the query.
@@ -70,6 +89,11 @@ class Replay( object ):
             if 0 == self._sensorPending:
                 isSingleQuery = True
                 self._sensorPending = 1
+                # Only initialize the limits if we are not called for
+                # an org-wide query. Otherwise the caller will have
+                # set the limits already.
+                self._limitEvent = limitEvent
+                self._limitEval = limitEval
 
         try:
             # Split up the total time into windows we can query in parallel.
@@ -85,7 +109,21 @@ class Replay( object ):
             with self._statusMutex:
                 self._queryPending += len( windows )
 
-            results = parallelExec( lambda w: self._scanHistoricalSensor( sid, w[ 0 ], w[ 1 ], ruleName = ruleName, ruleContent = ruleContent, isRunTrace = isRunTrace ), windows, maxConcurrent = self._maxConcurrent )
+            if isIgnoreState:
+                # Just run it all together.
+                results = parallelExec( lambda w: self._scanHistoricalSensor( sid, w[ 0 ], w[ 1 ], ruleName = ruleName, ruleContent = ruleContent, isRunTrace = isRunTrace, isIgnoreState = isIgnoreState ), windows, maxConcurrent = self._maxConcurrent )
+            else:
+                # We need to iterate one at a time and keep passing the state along.
+                results = []
+                state = None
+                for w in windows:
+                    tmpResult = self._scanHistoricalSensor( sid, w[ 0 ], w[ 1 ], ruleName = ruleName, ruleContent = ruleContent, isRunTrace = isRunTrace, isIgnoreState = isIgnoreState, state = state )
+                    if 'state' in tmpResult:
+                        # Extract the state update so we can pass it to the next call.
+                        state = tmpResult[ 'state' ]
+                        # Remove the state data from the result.
+                        del( tmpResult[ 'state' ] )
+                    results.append( tmpResult )
         finally:
             with self._statusMutex:
                 self._sensorPending -= 1
@@ -97,10 +135,12 @@ class Replay( object ):
 
         return self._rollupResults( results )
 
-    def _scanHistoricalSensor( self, sid, startTime, endTime, ruleName = None, ruleContent = None, isRunTrace = False ):
+    def _scanHistoricalSensor( self, sid, startTime, endTime, ruleName = None, ruleContent = None, isRunTrace = False, isIgnoreState = False, state = None ):
         # print( "Starting query %s-%s for %s" % ( startTime, endTime, sid ) )
         # qStart = time.time()
 
+        resp = None
+
         try:
             if self._apiURL is None:
                 # Get the ingest URL from the API.
@@ -111,14 +151,35 @@ class Replay( object ):
             }
             if isRunTrace:
                 req[ 'trace' ] = 'true'
-            body = None
+            if isIgnoreState is False:
+                req[ 'return_state' ] = 'true'
+
+            # If limits were reached, short circuit the execution.
+            with self._statusMutex:
+                if self._limitEvent is not None:
+                    if 0 >= self._limitEvent:
+                        return {}
+                    req[ 'limit_event' ] = str( self._limitEvent )
+                if self._limitEval is not None:
+                    if 0 >= self._limitEval:
+                        return {}
+                    req[ 'limit_eval' ] = str( self._limitEval )
+
+            body = {}
+            if state is not None:
+                body[ 'state' ] = state
             if ruleName is not None:
                 req[ 'rule_name' ] = ruleName
             elif ruleContent is not None:
-                body = json.dumps( ruleContent )
+                body[ 'rule' ] = ruleContent
             else:
                 raise LcApiException( 'no rule specified' )
 
+            if 0 == len( body ):
+                body = None
+            else:
+                body = json.dumps( body ).encode()
+
             nRetry = 0
             while True:
                 try:
@@ -139,11 +200,18 @@ class Replay( object ):
         finally:
             with self._statusMutex:
                 self._queryPending -= 1
+
+                # If there are limits, compute the left-over quota.
+                if isinstance( resp, dict ):
+                    if self._limitEvent is not None:
+                        self._limitEvent -= resp.get( 'num_events', 0 )
+                    if self._limitEval is not None:
+                        self._limitEval -= resp.get( 'num_evals', 0 )
         # print( "Finished query %s-%s for %s in %s seconds" % ( startTime, endTime, sid, time.time() - qStart ) )
 
         return resp
 
-    def scanEntireOrg( self, startTime, endTime, ruleName = None, ruleContent = None, isRunTrace = False ):
+    def scanEntireOrg( self, startTime, endTime, ruleName = None, ruleContent = None, isRunTrace = False, limitEvent = None, limitEval = None, isIgnoreState = False ):
         '''Scan an entire organization's data with a D&R rule.
 
         Args:
@@ -152,21 +220,21 @@ class Replay( object ):
             ruleName (str): the name of an existing D&R rule to use.
             ruleContent (dict): D&R rule to use to scan, with a "detect" key and a "respond" key.
             isRunTrace (bool): if True, generate a trace of the evaluation.
+            limitEvent (int): approximately limit the number of events evaluated.
+            limitEval (int): approximately limit the number of rule evaluations.
+            isIgnoreState (bool): if True, parallelize processing of single sensors to increase performance but limit effectiveness of stateful detection.
 
         Returns:
             a dict containing results of the query.
         '''
-        sensors = self._lc.sensors()
-        while True:
-            moreSensors = self._lc.sensors( is_next = True )
-            if moreSensors is None:
-                break
-            sensors += moreSensors
+        sensors = list( self._lc.sensors() )
 
         with self._statusMutex:
             self._sensorPending = len( sensors )
+            self._limitEvent = limitEvent
+            self._limitEval = limitEval
 
-        results = parallelExec( lambda sid: self.scanHistoricalSensor( sid, startTime, endTime, ruleName = ruleName, ruleContent = ruleContent, isRunTrace = isRunTrace ), sensors, maxConcurrent = self._maxConcurrent )
+        results = parallelExec( lambda sid: self.scanHistoricalSensor( sid, startTime, endTime, ruleName = ruleName, ruleContent = ruleContent, isRunTrace = isRunTrace, isIgnoreState = isIgnoreState ), sensors, maxConcurrent = self._maxConcurrent )
 
         with self._statusMutex:
             if self._queryStartedAt is not None:
@@ -174,7 +242,7 @@ class Replay( object ):
 
         return self._rollupResults( results )
 
-    def scanEvents( self, events, ruleName = None, ruleContent = None, isRunTrace = False ):
+    def scanEvents( self, events, ruleName = None, ruleContent = None, isRunTrace = False, limitEvent = None, limitEval = None ):
         '''Scan the specific events with a D&R rule.
 
         Args:
@@ -182,6 +250,8 @@ class Replay( object ):
             ruleName (str): the name of an existing D&R rule to use.
             ruleContent (dict): D&R rule to use to scan, with a "detect" key and a "respond" key.
             isRunTrace (bool): if True, generate a trace of the evaluation.
+            limitEvent (int): approximately limit the number of events evaluated.
+            limitEval (int): approximately limit the number of rule evaluations.
 
         Returns:
             a dict containing results of the query.
@@ -209,6 +279,11 @@ class Replay( object ):
             if isRunTrace:
                 req[ 'trace' ] = 'true'
 
+            if limitEvent is not None:
+                req[ 'limit_event' ] = str( limitEvent )
+            if limitEval is not None:
+                req[ 'limit_eval' ] = str( limitEval )
+
             nRetry = 0
             while True:
                 try:
@@ -233,13 +308,62 @@ class Replay( object ):
 
         return resp
 
+    def validateRule( self, ruleContent = None ):
+        '''Validate a D&R rule compiles properly.
+
+        Args:
+            ruleContent (dict): D&R rule to use to scan, with a "detect" key and a "respond" key.
+
+        Returns:
+            a dict containing results of the query.
+        '''
+        # qStart = time.time()
+
+        if self._apiURL is None:
+            # Get the ingest URL from the API.
+            self._apiURL = 'https://%s/' % ( self._replayURL, )
+        req = {}
+        body = {
+            'rule' : ruleContent,
+        }
+        if ruleContent is None:
+            raise LcApiException( 'no rule specified' )
+
+        nRetry = 0
+        while True:
+            try:
+                resp = self._lc._apiCall( 'validate/%s' % ( self._lc._oid, ),
+                                          'POST',
+                                          {},
+                                          altRoot = self._apiURL,
+                                          queryParams = req,
+                                          rawBody = json.dumps( body ).encode(),
+                                          contentType = 'application/json' )
+
+                break
+            except:
+                nRetry += 1
+                if nRetry > 5:
+                    raise
+                time.sleep( 2 * nRetry )
+        # print( "Finished query %s-%s for %s in %s seconds" % ( startTime, endTime, sid, time.time() - qStart ) )
+
+        return resp
+
     def _rollupResults( self, results ):
         final = {}
         for result in results:
             if not isinstance( result, dict ):
                 raise result
             for k, v in result.items():
-                if isinstance( v, ( int, float ) ):
+                if isinstance( v, bool ):
+                    # MUST test for bool before int because in Python
+                    # a bool IS an instance of an INT.....
+                    if k not in final:
+                        final[ k ] = v
+                    elif not final[ k ] and v:
+                        final[ k ] = v
+                elif isinstance( v, ( int, float ) ):
                     if k not in final:
                         final[ k ] = 0
                     final[ k ] += v
@@ -254,7 +378,7 @@ class Replay( object ):
                     raise LcApiException( 'unexpected data type: %s' % ( type( v ), ) )
         return final
 
-def main():
+def main( sourceArgs = None ):
     import argparse
 
     parser = argparse.ArgumentParser( prog = 'limacharlie.io replay detection and response' )
@@ -343,7 +467,35 @@ def main():
                          dest = 'isRunTrace',
                          help = 'if set will output a trace of each operator evaluation and the result' )
 
-    args = parser.parse_args()
+    parser.add_argument( '--limit-event',
+                         type = int,
+                         required = False,
+                         dest = 'limitEvent',
+                         default = None,
+                         help = 'limits the number of events evaluated to approximately this number.' )
+
+    parser.add_argument( '--limit-eval',
+                         type = int,
+                         required = False,
+                         dest = 'limitEval',
+                         default = None,
+                         help = 'limits the number of rule evaluations to approximately this number.' )
+
+    parser.add_argument( '--validate',
+                         action = 'store_true',
+                         default = False,
+                         required = False,
+                         dest = 'isValidate',
+                         help = 'if set will only validate the rule compiles properly' )
+
+    parser.add_argument( '--ignore-state',
+                         action = 'store_true',
+                         default = False,
+                         required = False,
+                         dest = 'isIgnoreState',
+                         help = 'if set, processing from single sensors will be parallelized increasing performance but limiting effectiveness of stateful detection.' )
+
+    args = parser.parse_args( sourceArgs )
 
     replay = Replay( Manager( None, None ),
                      isInteractive = args.isInteractive,
@@ -362,57 +514,79 @@ def main():
             except:
                 raise LcApiException( 'rule content not valid yaml or json' )
 
-    if args.events is None:
-        if ( args.start is None or args.end is None ) and args.lastSeconds is None:
-            raise LcApiException( 'must specify start and end, or last-seconds' )
-
-    if args.events is None:
-        # We want to use Insight-based events.
-        start = args.start
-        end = args.end
-        if start is None and end is None and args.lastSeconds is not None:
-            now = int( time.time() )
-            start = now - args.lastSeconds
-            end = now
-
-        if args.sid is not None:
-            response = replay.scanHistoricalSensor( str( args.sid ),
-                                                    start,
-                                                    end,
-                                                    ruleName = args.ruleName,
-                                                    ruleContent = ruleContent,
-                                                    isRunTrace = args.isRunTrace )
-        elif args.isEntireOrg:
-            response = replay.scanEntireOrg( start,
-                                             end,
-                                             ruleName = args.ruleName,
-                                             ruleContent = ruleContent,
-                                             isRunTrace = args.isRunTrace )
-        else:
-            raise LcApiException( '--sid or --entire-org must be specified' )
+    if args.isValidate:
+        response = replay.validateRule( ruleContent )
     else:
-        # We are using an events file.
-        with open( args.events, 'rb' ) as f:
-            fileContent = f.read().decode()
-        # We support two formats.
-        try:
-            if "\n" in fileContent and not fileContent.startswith( "[" ):
-                # This is newline-delimited like you get from LC Outputs.
-                events = [ json.loads( e ) for e in fileContent.split( '\n' ) ]
+        if args.events is None:
+            if ( args.start is None or args.end is None ) and args.lastSeconds is None:
+                print( 'must specify start and end, or last-seconds' )
+                parser.print_help()
+                sys.exit(1)
+
+        if args.events is None:
+            # We want to use Insight-based events.
+            start = args.start
+            end = args.end
+            if start is None and end is None and args.lastSeconds is not None:
+                now = int( time.time() )
+                start = now - args.lastSeconds
+                end = now
+
+            if args.sid is not None:
+                response = replay.scanHistoricalSensor( str( args.sid ),
+                                                        start,
+                                                        end,
+                                                        ruleName = args.ruleName,
+                                                        ruleContent = ruleContent,
+                                                        isRunTrace = args.isRunTrace,
+                                                        limitEvent = args.limitEvent,
+                                                        limitEval = args.limitEval,
+                                                        isIgnoreState = args.isIgnoreState )
+            elif args.isEntireOrg:
+                response = replay.scanEntireOrg( start,
+                                                 end,
+                                                 ruleName = args.ruleName,
+                                                 ruleContent = ruleContent,
+                                                 isRunTrace = args.isRunTrace,
+                                                 limitEvent = args.limitEvent,
+                                                 limitEval = args.limitEval,
+                                                 isIgnoreState = args.isIgnoreState )
             else:
-                # This is a JSON list containing all the events like you get
-                # from the historical view download button.
-                events = json.loads( fileContent )
-        except:
-            print( "!!! Invalid events provided. Content should be a JSON LIST of events or newline-separated JSON." )
-            sys.exit( 1 )
-        response = replay.scanEvents( events,
-                                      ruleName = args.ruleName,
-                                      ruleContent = ruleContent,
-                                      isRunTrace = args.isRunTrace )
+                raise LcApiException( '--sid or --entire-org must be specified' )
+        else:
+            # We are using an events file.
+            with open( args.events, 'rb' ) as f:
+                fileContent = f.read().decode()
+            # We support two formats.
+            try:
+                try:
+                    # This is a JSON list containing all the events like you get
+                    # from the historical view download button. Or just single
+                    # JSON event.
+                    events = json.loads( fileContent )
+                except:
+                    # This is newline-delimited like you get from LC Outputs.
+                    events = [ json.loads( e ) for e in fileContent.split( '\n' ) ]
+
+                # If the result is a dictionary and not a list we assume this was
+                # just a single event so we will wrap it.
+                if isinstance( events, dict ):
+                    events = [ events ]
+            except:
+                print( "!!! Invalid events provided. Content should be a JSON event, a JSON LIST of events or newline-separated JSON." )
+                sys.exit( 1 )
+            response = replay.scanEvents( events,
+                                          ruleName = args.ruleName,
+                                          ruleContent = ruleContent,
+                                          isRunTrace = args.isRunTrace,
+                                          limitEvent = args.limitEvent,
+                                          limitEval = args.limitEval )
 
     if args.isInteractive:
         # If this is interactive, we displayed progress, so we need a new line.
         print( "" )
 
-    print( json.dumps( response, indent = 2 ) )
\ No newline at end of file
+    print( json.dumps( response, indent = 2 ) )
+
+if '__main__' == __name__:
+    main()
\ No newline at end of file
