@@ -1,8 +1,13 @@
 import uuid
-import json
 import time
+import json
 
-from .utils import *
+from .utils import LcApiException
+from .utils import GET
+from .utils import DELETE
+from .utils import POST
+from .utils import FutureResults
+from .utils import enhanceEvent
 
 import gevent
 
@@ -14,9 +19,14 @@ class Sensor( object ):
     _PLATFORM_MACOS = 0x30000000
     _PLATFORM_IOS = 0x40000000
     _PLATFORM_ANDROID = 0x50000000
+    _PLATFORM_CHROMEOS = 0x60000000
 
     _ARCHITECTURE_X86 = 0x00000001
     _ARCHITECTURE_X64 = 0x00000002
+    _ARCHITECTURE_ARM = 0x00000003
+    _ARCHITECTURE_ARM64 = 0x00000004
+    _ARCHITECTURE_ALPINE64 = 0x00000005
+    _ARCHITECTURE_CHROME = 0x00000006
 
     def __init__( self, manager, sid ):
         try:
@@ -43,9 +53,11 @@ class Sensor( object ):
     def waitToComeOnline( self, timeout ):
         '''Wait for the sensor to be online.
 
-        :param timeout: number of seconds to wait up to
+        Args:
+            timeout (int): number of seconds to wait up to
 
-        :return True if sensor is back or False if timeout
+        Returns:
+            True if sensor is back or False if timeout
         '''
         deadline = time.time() + timeout
 
@@ -206,19 +218,26 @@ class Sensor( object ):
             self._PLATFORM_MACOS : 'macos',
             self._PLATFORM_IOS : 'ios',
             self._PLATFORM_ANDROID : 'android',
+            self._PLATFORM_CHROMEOS : 'chromeos',
         }
         archToString = {
             self._ARCHITECTURE_X86 : 'x86',
             self._ARCHITECTURE_X64 : 'x64',
+            self._ARCHITECTURE_ARM : 'arm',
+            self._ARCHITECTURE_ARM64 : 'arm64',
+            self._ARCHITECTURE_ALPINE64 : 'alpine64',
+            self._ARCHITECTURE_CHROME : 'chrome',
         }
         data = data[ 'info' ]
 
         self._platform = data[ 'plat' ]
         self._architecture = data[ 'arch' ]
         self._hostname = data.get( 'hostname', None )
+        self._is_isolated = data.get( 'is_isolated', None )
+        self._should_isolate = data.get( 'should_isolate', None )
 
-        data[ 'plat' ] = platToString[ data[ 'plat' ] ]
-        data[ 'arch' ] = archToString[ data[ 'arch' ] ]
+        data[ 'plat' ] = platToString.get( data[ 'plat' ], data[ 'plat' ] )
+        data[ 'arch' ] = archToString.get( data[ 'arch' ], data[ 'arch' ] )
 
         return data
 
@@ -266,6 +285,28 @@ class Sensor( object ):
             self.getInfo()
         return self._platform == self._PLATFORM_LINUX
 
+    def isChrome( self ):
+        '''Checks if the sensor is on Chrome.
+
+        Returns:
+            True if the sensor is Chrome.
+        '''
+        if self._architecture is None:
+            # If the platform has not been cached yet, retrieve the info.
+            self.getInfo()
+        return self._architecture == self._ARCHITECTURE_CHROME
+
+    def isChromeOS( self ):
+        '''Checks if the sensor is on ChromeOS.
+
+        Returns:
+            True if the sensor is on ChromeOS.
+        '''
+        if self._architecture is None:
+            # If the platform has not been cached yet, retrieve the info.
+            self.getInfo()
+        return self._platform == self._PLATFORM_CHROMEOS
+
     def hostname( self ):
         '''Get the hostname of this sensor.
 
@@ -277,7 +318,7 @@ class Sensor( object ):
             self.getInfo()
         return self._hostname
 
-    def getHistoricEvents( self, start, end, limit = None, eventType = None ):
+    def getHistoricEvents( self, start, end, limit = None, eventType = None, isForward = True ):
         '''Get the events for this sensor between the two times, requires Insight (retention) enabled.
 
         Args:
@@ -285,10 +326,12 @@ class Sensor( object ):
             end (int): end unix (seconds) timestamp to feth events to.
             limit (int): maximum number of events to return.
             eventType (str): return events only of this type.
+            isForward (bool): return events in ascending order.
 
         Returns:
-            a list of events.
+            a generator of events.
         '''
+        cursor = '-'
         start = int( start )
         end = int( end )
         if limit is not None:
@@ -298,6 +341,7 @@ class Sensor( object ):
             'start' : start,
             'end' : end,
             'is_compressed' : 'true',
+            'is_forward' : 'true' if isForward else 'false',
         }
 
         if limit is not None:
@@ -306,8 +350,18 @@ class Sensor( object ):
         if eventType is not None:
             req[ 'event_type' ] = eventType
 
-        data = self._manager._apiCall( 'insight/%s/%s' % ( self._manager._oid, self.sid ), GET, queryParams = req )
-        return [ enhanceEvent( e ) for e in self._manager._unwrap( data[ 'events' ] ) ]
+        nReturned = 0
+        while cursor:
+            req[ 'cursor' ] = cursor
+            data = self._manager._apiCall( 'insight/%s/%s' % ( self._manager._oid, self.sid ), GET, queryParams = req )
+            cursor = data.get( 'next_cursor', None )
+            for event in self._manager._unwrap( data[ 'events' ] ):
+                yield enhanceEvent( event )
+                nReturned += 1
+                if limit is not None and limit <= nReturned:
+                    break
+            if limit is not None and limit <= nReturned:
+                break
 
     def getHistoricOverview( self, start, end ):
         '''Get a list of timestamps representing where sensor data is available in Insight (retention).
@@ -344,6 +398,63 @@ class Sensor( object ):
         batches = self.getHistoricOverview( timestamp, timestamp + ( 60 * 60 * 2 ) )
         return 0 != len( batches )
 
+    def getObjectTimeline( self, start, end, bucketing = 'day', onlyTypes = None ):
+        '''Get summarized information about timeline of Objects (IOCs) for this host.
+
+        Args:
+            start (int): start time (unix seconds epoch) of the period to search.
+            end (int): end time (unix seconds epoch) of the period to search.
+            bucketing (str): granularity of the timeline, one of "hour", "day", "week", "month".
+            onlyTypes (list): list of object types to look for, all if undefined.
+
+        Returns:
+            Dict of timelines per type and object.
+        '''
+        req = {
+            'oid' : self._manager._oid,
+            'start' : start,
+            'end' : end,
+            'bucketing' : bucketing,
+            'sid' : self.sid,
+            'is_compressed' : True,
+            'objects' : json.dumps( { t : [] for t in onlyTypes } ) if onlyTypes is not None else "{}",
+        }
+
+        data = self._manager._apiCall( 'insight/%s/objects_timeline' % ( self._manager._oid, ), POST, req )
+        data = self._manager._unwrap( data[ 'timeline' ] )
+
+        # We don't care about the prevalence counters in this case
+        # so we'll just strip it down.
+        for oType, objects in data.items():
+            for o, timeInfo in objects.items():
+                objects[ o ] = sorted( tuple( timeInfo.keys() ) )
+
+        return data
+
+    def delete( self ):
+        '''Delete the sensor. It will not be able to connect to the cloud anymore, but will not be uninstalled.abs
+        '''
+
+        return self._manager._apiCall( '%s' % self.sid, DELETE, {} )
+
+    def isIsolatedFromNetwork( self ):
+        '''Determine if the given sensor is marked to be isolated from the network.
+
+        Returns:
+            True if isolated.
+        '''
+        # Network isolation is ephemeral, so always refresh.
+        self.getInfo()
+        return self._should_isolate
+
+    def isolateNetwork( self ):
+        '''Mark the sensor for network isolation (persistent).'''
+        return self._manager._apiCall( '%s/isolation' % ( self.sid, ), POST )
+
+    def rejoinNetwork( self ):
+        '''Remove the sensor from network isolation (persistent).'''
+        return self._manager._apiCall( '%s/isolation' % ( self.sid, ), DELETE )
+
     def __str__( self ):
         return self.sid
 
